@@ -7,14 +7,16 @@ from functools import partial
 import traceback
 from contextlib import redirect_stdout, redirect_stderr, ExitStack
 import sys
+import logging
+logger = logging.getLogger(__name__)
 
 from .contextvar import GVAR
-from .util import timeblock, WaitGroup, nullacontext, optional_cm, make_prependable
+from .util import timeblock, WaitGroup, nullacontext, optional_cm, make_prependable, load_contextmod
 from .resource import ZERO, Reject
 from .pair import SocketChannel, SocketChannelAcceptor
 from .subproc import subprocess_pool_daemon, BrokenSubprocessError
 from .stat import log_event, log_timestart, log_timeend, stat_logger, ls_worker_res, reg_probe, collect_probe
-from .config import load_conf
+from .config import load_conf, override
 
 FULL_HEALTH = 3
 STATSPAN = 15 # minutes
@@ -34,31 +36,32 @@ class GrainRemote(object):
     async def _loop(self, task_status=trio.TASK_STATUS_IGNORED):
         async def heartbeat_s(c):
             while True:
-                await c.try_send(b"HBT")
+                await c.try_send(dict(cmd="HBT"))
                 await trio.sleep(HEARTBEAT_INTERVAL)
-        async def heartbeat_n_receive(c):
-            async with trio.open_nursery() as _n:
-                _n.start_soon(heartbeat_s, c)
-                while True:
-                    with trio.move_on_after(HEARTBEAT_INTERVAL*HEARTBEAT_TOLERANCE):
-                        async for x in c:
-                            if x == b"HBT": break
-                            yield x
-                        else: break
-                        continue
-                    print(f"remote {self.name} heartbeat response timeout")
-                    break
-                _n.cancel_scope.cancel()
-        async with self._c:
+        async with self._c, \
+                   trio.open_nursery() as _n:
             task_status.started()
-            async for x in heartbeat_n_receive(self._c):
-                tid, r = pickle_loads(x)
-                rq = self.resultq.get(abs(tid))
+            _n.start_soon(heartbeat_s, self._c)
+            while True:
+                x = None
+                try:
+                    with trio.move_on_after(HEARTBEAT_INTERVAL*HEARTBEAT_TOLERANCE):
+                        x = await self._c.receive()
+                except trio.EndOfChannel:
+                    break
+                if x is None:
+                    logger.warning(f"remote {self.name} heartbeat response timeout")
+                    break
+                elif x == {'cmd':'HBT'}:
+                    continue
+                tid, ok, r = x['tid'], x['ok'], x['result']
+                rq = self.resultq.get(tid)
                 if rq:
-                    await rq.send((tid>0, r))
+                    await rq.send((ok, pickle_loads(r)))
                 else:
                     log_event("late_response")
-                    print(f"remote {self.name} received phantom job {abs(tid)}'s result")
+                    logger.warning(f"remote {self.name} received phantom job {tid}'s result")
+            _n.cancel_scope.cancel()
         # well, we just let the remote decide when to leave
         #assert self._c.is_clean is False # NOTE: P2P conn doesn't get EOF when the other end quit, we need to close on our end
         # In case of connection lost, dismiss all pending jobs
@@ -71,7 +74,7 @@ class GrainRemote(object):
         with self.wg:
             self.resultq[tid], rq = trio.open_memory_channel(0)
             try:
-                await self._c.send(pickle_dumps((tid, res, fn)))
+                await self._c.send(dict(tid=tid, res=res, func=pickle_dumps(fn)))
                 with optional_cm(trio.fail_after, getattr(res,'T',-180)+180): # 3min grace period
                     ok, r = await rq.receive()
             except (trio.ClosedResourceError, trio.EndOfChannel):
@@ -84,11 +87,11 @@ class GrainRemote(object):
             del self.resultq[tid]
             return ok, r
     async def aclose(self):
-        await self._c.try_send(b"FIN") # fails if KI or connection lost
-        print(f"worker {self.name} starts cleaning up")
+        await self._c.try_send(dict(cmd="FIN")) # fails if KI or connection lost
+        logger.info(f"worker {self.name} starts cleaning up")
         await self.wg.wait()
         await self._c.aclose() # for P2P connection
-        print(f"worker {self.name} clear")
+        logger.info(f"worker {self.name} clear")
 
 class GrainReverseRemote(GrainRemote):
     """A Grain remote for which the worker dial in
@@ -113,7 +116,7 @@ class _cobj:
             return getattr(self._h, attr)
         return getattr(self._dh, attr)
 class GrainPseudoRemote:
-    def __init__(self, res, out=""):
+    def __init__(self, res, out="", cmod=""):
         self.res = res
         self.name = f"{trio.socket.gethostname()}(local)"
         self.health = FULL_HEALTH
@@ -124,13 +127,20 @@ class GrainPseudoRemote:
             outh = self.redi_cm.enter_context(open(out,'w'))
             self.redi_cm.enter_context(redirect_stdout(_cobj(outh,sys.stdout)))
             self.redi_cm.enter_context(redirect_stderr(_cobj(outh,sys.stderr)))
+        self.cmod = cmod
+    async def _scope(self, task_status=trio.TASK_STATUS_IGNORED):
+        with trio.CancelScope() as cs, \
+             self.redi_cm:
+            self.cg.add(cs)
+            async with load_contextmod(self.cmod)():
+                task_status.started()
+                await trio.sleep_forever()
     async def connect(self, _n):
-        if self.res > ZERO:
-            self.cg.add(await _n.start(subprocess_pool_daemon))
+        await _n.start(self._scope)
     async def execf(self, tid, res, fn):
-        cs = trio.CancelScope()
-        self.cg.add(cs)
-        with self.wg, cs:
+        with self.wg, \
+             trio.CancelScope() as cs:
+            self.cg.add(cs)
             GVAR.res = res
             GVAR.instance = "local"
             try:
@@ -145,10 +155,9 @@ class GrainPseudoRemote:
     async def aclose(self):
         for cs in self.cg:
             cs.cancel()
-        print(f"worker {self.name} starts cleaning up")
+        logger.info(f"worker {self.name} starts cleaning up")
         await self.wg.wait()
-        self.redi_cm.__exit__(None, None, None)
-        print(f"worker {self.name} clear")
+        logger.info(f"worker {self.name} clear")
 
 
 class GrainManager:
@@ -175,9 +184,9 @@ class GrainManager:
             w.health -= 1 if type(err) in self.temperr else INFIN
             if w.health <= 0 and w in self.pool:
                 if type(err) not in self.temperr:
-                    print(tb)
+                    logger.error(tb)
                 if w.name.endswith('(local)'): raise RuntimeError("local worker quits")
-                print(f"quit worker {w.name} due to poor health {w.health}")
+                logger.warning(f"quit worker {w.name} due to poor health {w.health}")
                 await self.unregister(w.name, locked=True)
 
     async def dealloc(self, w, res):
@@ -220,31 +229,31 @@ class GrainManager:
                     msg = await _c.receive()
                 except trio.EndOfChannel:
                     continue
-                cmd, msg = msg[:3], msg[3:]
-                if cmd == b"CON": # connect
-                    vaddr, res = pickle_loads(msg)
-                    print(f"worker {vaddr} joined with {res}")
+                cmd = msg['cmd']
+                if cmd == "CON": # connect
+                    vaddr, res = msg['name'], msg['res']
+                    logger.info(f"worker {vaddr} joined with {res}")
                     await self.register(GrainReverseRemote(_c, vaddr, res), _n)
                     continue
                 async with _c: # The following are ephemeral cmds
-                    if cmd == b"REG": # register
-                        addr, res = pickle_loads(msg)
-                        print(f"worker {addr} joined with {res}")
+                    if cmd == "REG": # register
+                        addr, res = msg['name'], msg['res']
+                        logger.info(f"worker {addr} joined with {res}")
                         await self.register(GrainRemote(addr, res), _n)
-                    elif cmd == b"UNR": # unregister
-                        addr = msg.decode()
-                        print(f"worker {addr} asked for quit")
+                    elif cmd == "UNR": # unregister
+                        addr = msg['name']
+                        logger.info(f"worker {addr} is quitting now")
                         await self.unregister(addr)
-                    elif cmd == b"STA": # statistics
+                    elif cmd == "STA": # statistics
                         sta = ls_worker_res(self.pool) + '\n' + \
                               collect_probe()
-                        await _c.try_send(sta.encode())
-                    elif cmd == b"TRM": # graceful termination
-                        addr = msg.decode()
-                        print(f"worker {addr} is going to leave")
+                        await _c.try_send(dict(result=sta.encode()))
+                    elif cmd == "TRM": # graceful termination
+                        addr = msg['name']
+                        logger.info(f"worker {addr} is going to leave")
                         _n.start_soon(self.terminate, addr)
                     else:
-                        print(f"worker manager received unknown command {cmd} from {_c.host}")
+                        logger.warning(f"worker manager received unknown command {cmd} from {_c.host}")
 
     async def __aenter__(self):
         await self._n.start(self.worker_manager)
@@ -286,11 +295,12 @@ class GrainExecutor(object):
         self._n = _n
         self._wg = WaitGroup() # track the entire lifetime of each job
         self.reschedule = reschedule
-        conf_head = load_conf(config_file).head
+        conf = load_conf(config_file)
+        override(conf, conf.head)
         self.mgr = GrainManager(
-            [GrainPseudoRemote(deepcopy(rpw if not nolocal else ZERO), conf_head.log_file)] + \
+            [GrainPseudoRemote(deepcopy(rpw if not nolocal else ZERO), conf.log_file, conf.contextmod)] + \
             [GrainRemote(a, deepcopy(rpw)) for a in waddrs],
-            self._n, temporary_err, persistent, conf_head.listen)
+            self._n, temporary_err, persistent, conf.listen)
         reg_probe("queued jobs", lambda: len(self.push_job._state.data))
         self.stat_tag = stat_tag
 
@@ -322,7 +332,7 @@ class GrainExecutor(object):
                 tb, err = r
                 if not self.reschedule:
                     raise RuntimeError(f"worker {w.name}'s task {fn} raises {err.__class__.__name__}: {err}, abort.")
-                print(f"worker {w.name}'s task {fn} raises {err.__class__.__name__}: {err}, going to reschedule it...")
+                logger.error(f"worker {w.name}'s task {fn} raises {err.__class__.__name__}: {err}, going to reschedule it...")
                 await self.mgr.health_check(w, tb, err)
                 self.resubmit(tid, res, fn) # preserve tid
             await self.mgr.dealloc(w, res)
