@@ -4,7 +4,8 @@ from math import inf as INFIN
 from functools import partial
 from urllib.parse import urlparse, parse_qsl
 
-from .conn import iter_packet, send_packet, open_tcp_stream_to_head, serve_tcp_p2p
+from .conn import open_tcp_stream_to_head, serve_tcp_p2p
+from .conn_msgp import iter_packet, send_packet
 
 async def notify(url, msg, seg=False): # TODO: retry until connected
     """Open a connection, send `msg`, then close the
@@ -112,7 +113,9 @@ class SocketChannelAcceptor(SocketReceiveChannel):
                                     handler_nursery=self._n, opts=self.opts, cs=self._cs))
         return self
     async def _handler(self, so):
-        host, port = so.socket.getpeername()
+        host, port, *_ = so.socket.getpeername()
+        if so.socket.proto == 6:
+            host = f"[{host}]"
         c = SocketChannel(f"tcp://{host}:{port}", _so=so)
         self.in_s.send_nowait(c)
         await c._handler_standalone(so) # to have the same lifetime as c
