@@ -1,7 +1,9 @@
-from contextlib import ContextDecorator
+from contextlib import ContextDecorator, asynccontextmanager
 from timeit import default_timer as timer
 from functools import wraps
 import types
+import importlib.util
+from pathlib import Path
 from math import inf as INF
 
 def timeblock(text="this block", enter=False):
@@ -20,7 +22,6 @@ def timeblock(text="this block", enter=False):
 import trio
 from trio.lowlevel import ParkingLot, checkpoint, enable_ki_protection
 from outcome import Value
-from async_generator import asynccontextmanager
 
 class WaitGroup(object):
 
@@ -185,6 +186,15 @@ def optional_cm(cm, cond_arg):
         return cm(cond_arg)
     return nullacontext()
 
+def load_contextmod(modp):
+    if modp == "":
+        return nullacontext
+    p = Path(modp)
+    spec = importlib.util.spec_from_file_location(p.stem, p)
+    mod = importlib.util.module_from_spec(spec)
+    spec.loader.exec_module(mod)
+    return mod.grain_context
+
 def set_numpy_oneline_repr():
     """Change the default Numpy array ``__repr__`` to a
     compact one. This is useful for keeping the log
@@ -195,6 +205,8 @@ def set_numpy_oneline_repr():
     EDGEITEMS = 3
     def oneline_repr(a):
         N = np.prod(a.shape)
+        if N == 0: # No numbers to show if the last dimension is zero
+            return f"array[{a.shape}]([])"
         ind = np.unravel_index(range(min(EDGEITEMS, N)), a.shape)
         afew = []
         for x,val in zip(np.c_[ind], a[ind]):
